# General purpose code for being a tray applet

proc manyset {list args} {
    foreach val $list var $args {
        upvar 1 $var my
        set my $val
    }
}


package require Tclx
package require tktray

#----- general machinery -----

# Interface:
#
#  tk::tktray widget is called .i
#
# Tooltip:
#
#   Caller may call
#      applet::setup-tooltip ON-VISIBLE ON-INVISIBLE
#   to make applet have a tooltip.
#
#   ON-VISIBLE and ON-INVISIBLE will be globally eval'd
#   when the tooltip becomes visible and invisible.
#
#   Caller should call
#      applet::tooltip-set TEXT-MAYBE-MULTILINE
#   whenever they like.
# 
# Button presses
#
#    Caller may bind .i.i <ButtonPress-$b>
#
#    Alternatively caller may call  applet::setup-button-menu $b
#    which will generate a menu .m$b which the user can configure
#    and which will automatically be posted and unposted etc.
#    In this case the caller should arrange that all of their
#    menus, when an item is selected, call
#      applet::msel
#
# Icon:
#
#  Caller should call:
#      applet::setimage IMAGE
#  as necessary.
#
# Alternatively, instead of icon, it may make other arrangements
# to use the provided subwindow
#
#  Caller that needs access to inner window should call
#      applet::setup-subwindow WIDTH ON-DESTROYING ON-READY
#  Then the main code will call ON-DESTROYING just before
#  destroying the inner window and recreating it, and ON-READY
#  just after.  In ON-READY the inner window is called .i.i.
#
#  The user of the subwindow machinery may call
#      applet::subwindow-need-recreate
#  if for any reason the inner window should be destroyed and recreated.
#
# Alternatively, 
# to run (repeatedly
# if necessary) a subprocess.
#     
#     proc innerwindow {} { ... }
#  and run
#     innerwindow
#  This will create
#     .i.i.b      frame to contain container
#     .i.i.b.c    actual container

wm withdraw .

tktray::icon .i -class example
.i configure -docked 1

fconfigure stdout -buffering line


namespace eval applet {


# used by both menus and tooltips
variable posted 0

#----- menus -----

proc setup-button-menu {b} {
    bind .i.i <ButtonPress> { applet::menubuttonpressed %b %X %Y }
    menu .m$b -tearoff 0
}

proc menubuttonpressed {b x y} {
    variable posted
    tooltip-cancel
    if {$posted == $b} {
	puts "unpost $posted toggle"
     	.m$posted unpost
	set posted 0
    } elseif {[winfo exists .m$b]} {
	if {$posted} {
	    .m$posted unpost
	    puts "unpost $posted other"
	}
	puts "post $b"
	set posted $b
	.m$b post $x $y
    }
}

proc msel {} {
    variable posted
    set posted 0
}

#----- tooltips -----

variable tooltip_on_vis {}
variable tooltip_on_invis {}

proc tooltip-starttimer {state x y} {
    variable tooltip_after
    variable posted
    variable tooltip_inwindow
    if {$state || $posted || !$tooltip_inwindow} { tooltip-cancel; return }
    catch { after cancel $tooltip_after }
    set tooltip_after [after 500 applet::tooltip-show $x $y]
}

proc tooltip-cancel {} {
    variable tooltip_after
    variable tooltip_on_invis
    catch { after cancel $tooltip_after }
    catch { unset $tooltip_after }
    wm withdraw .tt
    uplevel #0 $tooltip_on_invis
}

set tooltip_inwindow 0

proc tooltip-enter {state x y} {
    variable tooltip_inwindow
    set tooltip_inwindow 1
    tooltip-starttimer $state $x $y
}

proc tooltip-leave {} {
    variable tooltip_inwindow
    set tooltip_inwindow 0
    tooltip-cancel
}

proc setup-tooltip {on_vis on_invis} {
    foreach v {vis invis} {
	variable tooltip_on_$v [set on_$v]
    }
    bind .i <Enter> { applet::tooltip-enter %s %X %Y }
    bind .i <Leave> { applet::tooltip-leave }
    bind .i <ButtonRelease> { 
	applet::tooltip-cancel
	applet::tooltip-starttimer %s %X %Y
    }
    bind .i <Motion> { applet::tooltip-starttimer %s %X %Y }
    toplevel .tt -background black
    wm withdraw .tt
    wm overrideredirect .tt 1
    label .tt.t -justify left -background {#EEE1B3}
    pack .tt.t -padx 1 -pady 1
    tooltip-set {}
}

proc tooltip-set {s} {
    .tt.t configure -text $s
}

proc tooltip-show {x y} {
    variable tooltip_on_vis
    incr x 9
    incr y 9
    wm geometry .tt +$x+$y
    wm deiconify .tt
    uplevel #0 $tooltip_on_vis
}

#----- simple images -----

proc setimage {image} {
    .i configure -image $image
}

#----- subwindow -----

variable subwindow_on_destroying
variable subwindow_on_ready

proc subwindow-need-recreate {} {
    variable innerwindow_after
puts "IW-EVENT"
    if {[info exists innerwindow_after]} return
    set innerwindow_after [after idle applet::innerwindow-resetup]
}

proc innerwindow-resetup {} {
    variable innerwindow_after
    variable subwindow_on_destroying
    variable subwindow_on_ready
    unset innerwindow_after

puts RESETUP

    uplevel #0 $subwindow_on_destroying

    catch { destroy .i.i.c }
    if {![winfo exists .i.i]} return
    destroy [frame .i.i.make-exist]
    catch { destroy .i.i.b.c }
    catch { destroy .i.i.b }
    frame .i.i.b
    pack .i.i.b -fill both -side left -expand 1
    frame .i.i.b.c -container 1 -background orange
    pack .i.i.b.c -fill both -side left -expand 1
#
    global inner_lastw inner_lasth
    #set w [winfo width .i.i]
#    set w [winfo width .i.i]
#    set h [winfo height .i.i]

#    if {$w != $inner_lastw || $h != $inner_lasth} {
#	set inner_lastw $w
#	set inner_lasth $h
#	innerwindow-ph-dummy configure -width $w -height 2

    uplevel #0 $subwindow_on_ready
#    }
}

proc setup-subwindow {w on_destroying on_ready} {
    foreach v {destroying ready} {
	variable subwindow_on_$v [set on_$v]
    }

    global inner_lastw inner_lasth
    set inner_lastw -2
    set inner_lasth -2

    image create photo applet::innerwindow-ph-dummy -width $w -height 2
    .i configure -image applet::innerwindow-ph-dummy

    destroy [frame .i.make-exist]
    destroy [frame .i.i.make-exist]
    bind .i <<IconConfigure>> { 
	applet::subwindow-need-recreate
    }
}

#----- subprocess -----

variable subproc none
variable ratelimit {}

proc setup-subproc {w get_cmdline} {
    variable subproc_get_cmdline $get_cmdline
    setup-subwindow $w applet::subproc-destroying applet::subproc-ready
}

proc subproc-destroying {} {
    variable subproc
    puts "DESTROYING $subproc"
    switch -exact $subproc {
	none { }
	old { }
	default { kill $subproc; set subproc old }
    }
}

proc subproc-ready {} {
    variable subproc
    puts "READY $subproc"
    switch -exact $subproc {
	none {
	    run-child
	}
	old {
	    # wait for it to die
	}
	default {
	    error "unexpected state $subproc"
	}
    }
    puts "READY-done $subproc"
}

proc run-child {} {
    variable subproc
    variable ratelimit
    variable subproc_get_cmdline

    set id [winfo id .i.i.b.c]
    set cmd [uplevel #0 [concat [list $subproc_get_cmdline] $id]]

    puts "RUN-CHILD $subproc"
    set now [clock seconds]
    lappend ratelimit $now
    while {[lindex $ratelimit 0] < {$now - 10}} {
	set ratelimit [lrange $ratelimit 1 end]
    }
    if {[llength $ratelimit] > 10} {
	puts stderr "crashing repeatedly, quitting $ratelimit"
	exit 127
    }

    set subproc none
    set subproc [subproc::fork applet::child-died {
	execl [lindex $cmd 0] [lrange $cmd 1 end]
    }]
    puts "FORKED $subproc"
}

proc child-died {how how2} {
    puts "DIED $how $how2"
    variable subproc
    switch -exact $subproc {
	old {
	    set subproc none
	    run-child
	}
	default {
	    set subproc none
	    subwindow-need-recreate
	}
    }
}

}
